from typing import Tuple

from ecdsa import SECP256k1
from ecdsa.curves import Curve

from .coding import (encode_number,
                     hex_string_to_int)
from .coordinates import jacobian
from .coordinates.utils import modular_multiplicative_inverse
from .hashes import keccak_256_hash
from .restrictions import (R_LENGTH,
                           S_LENGTH)


def verifying_key_hex_bytes(signature: str,
                            *,
                            message: str,
                            encoding: str = 'utf-8') -> bytes:
    v, r, s = decode_signature(signature)

    prepended_message = ('\x19Ethereum Signed Message:\n'
                         + str(len(message))
                         + message)
    message_hash = (keccak_256_hash(prepended_message.encode(encoding))
                    .hexdigest())
    return verifying_key_hex_bytes_from_hash(message_hash,
                                             v=v,
                                             r=r,
                                             s=s)


def verifying_key_hex_bytes_from_hash(message_hash: str,
                                      *,
                                      v: int,
                                      r: int,
                                      s: int) -> bytes:
    left, right = verifying_key_pair(message_hash,
                                     v=v,
                                     r=r,
                                     s=s)
    result = (encode_number(left,
                            base=256,
                            min_length=32)
              + encode_number(right,
                              base=256,
                              min_length=32))

    if len(result) != 64:
        err_msg = ('Invalid public key, '
                   'should have length of 64')
        raise ValueError(err_msg)

    return result


def verifying_key_pair(message_hash: str,
                       *,
                       v: int,
                       r: int,
                       s: int,
                       curve: Curve = SECP256k1) -> Tuple[int, int]:
    if not (27 <= v <= 28):
        raise ValueError(f'"v" must be equal to 27 or 28, '
                         f'but found {v}')

    curve_fp = curve.curve
    p = curve_fp.p()
    a = curve_fp.a()
    b = curve_fp.b()
    n = curve.order
    g_x = curve.generator.x()
    g_y = curve.generator.y()

    x = r
    beta = pow((x ** 3 + a * x + b) % p, (p + 1) // 4, p)
    y = beta if v % 2 ^ beta % 2 else p - beta
    z = hex_string_to_int(message_hash)
    g_z = jacobian.multiply(point=(g_x, g_y, 1),
                            multiplier=(n - z) % n)
    xy = jacobian.multiply(point=(x, y, 1),
                           multiplier=s)
    q_r = jacobian.add(g_z, xy)
    q = jacobian.multiply(point=q_r,
                          multiplier=modular_multiplicative_inverse(r, n))
    return jacobian.to_affine(q)


def decode_signature(signature: str) -> Tuple[int, int, int]:
    r = hex_string_to_int(signature[:R_LENGTH])
    s = hex_string_to_int(signature[R_LENGTH:R_LENGTH + S_LENGTH])
    v = hex_string_to_int(signature[R_LENGTH + S_LENGTH:])
    if v in {0, 1}:
        v += 27
    return v, r, s



from typing import (Any)

from hypothesis import (Verbosity,
                        find,
                        settings)
from hypothesis.searchstrategy import SearchStrategy


def example(strategy: SearchStrategy) -> Any:
    return find(specifier=strategy,
                condition=lambda x: True,
                settings=settings(max_shrinks=0,
                                  max_iterations=10_000,
                                  database=None,
                                  verbosity=Verbosity.quiet))



import os
import pkgutil
import sys
from typing import List

from _pytest.config import Parser
from _pytest.python import Metafunc

base_dir = os.path.dirname(__file__)
sys.path.append(base_dir)


def explore_pytest_plugins(base_dir: str,
                           fixtures_pkg_name: str) -> List[str]:
    fixtures_pkg_path = os.path.join(base_dir,
                                     fixtures_pkg_name)
    return [f'{fixtures_pkg_name}.{name}'
            for _, name, _ in pkgutil.iter_modules([fixtures_pkg_path])]


pytest_plugins = explore_pytest_plugins(base_dir=base_dir,
                                        fixtures_pkg_name='fixtures')


def pytest_addoption(parser: Parser) -> None:
    parser.addoption('--repeat', action='store',
                     help='Number of times to repeat each test')


def pytest_generate_tests(metafunc: Metafunc) -> None:
    if metafunc.config.option.repeat is not None:
        count = int(metafunc.config.option.repeat)
        # We're going to duplicate these tests by parametrisation,
        # which requires that each test has a fixture to accept the parameter.
        # We can add a new fixture like so:
        metafunc.fixturenames.append('tmp_ct')
        # Now we parametrize. This is what happens when we do e.g.,
        # @pytest.mark.parametrize('tmp_ct', range(count))
        # def test_foo(): pass
        metafunc.parametrize('tmp_ct', range(count))



from ethereum_helpers import recover


def test_verifying_key_hex_bytes(message: str,
                                 signature: str,
                                 verifying_key_hex_bytes: bytes) -> None:
    assert recover.verifying_key_hex_bytes(signature,
                                           message=message
                                           ) == verifying_key_hex_bytes



def modular_multiplicative_inverse(number: int,
                                   modulus: int) -> int:
    if number == 0:
        return 0
    x_1, x_2 = 1, 0
    u, v = number, modulus
    while u > 1:
        q = v // u
        x_1, x_2 = x_2 - q * x_1, x_1
        remainder = v - q * u
        u, v = remainder, u
    return x_1 % modulus



from typing import Tuple

from ethereum_helpers.types import JacobianPointType

from . import jacobian

AffinePointType = Tuple[int, int]


def to_jacobian(point: AffinePointType) -> JacobianPointType:
    return point[0], point[1], 1


def multiply(*,
             point: AffinePointType,
             multiplier: int) -> AffinePointType:
    return jacobian.to_affine(jacobian.multiply(point=to_jacobian(point),
                                                multiplier=multiplier))



from functools import partial
from hashlib import sha3_256
from hmac import HMAC
from typing import (Union,
                    Type,
                    Callable,
                    Dict,
                    Tuple)

from Cryptodome.Cipher import AES
from Cryptodome.Util import Counter
from ecdsa import (SECP256k1,
                   SigningKey,
                   VerifyingKey)
from ecdsa.curves import Curve

from . import aes
from .coding import (decode_hex,
                     big_endian_to_int)
from .hashes import (scrypt_hash,
                     keccak_256_hash)

HashFunctionType = Callable[..., HMAC]
KeyType = Union[SigningKey,
                VerifyingKey]
KeyClsType = Union[Type[SigningKey],
                   Type[VerifyingKey]]


def hex_bytes_to_key(key_cls: KeyClsType,
                     hex_bytes: bytes,
                     *,
                     curve: Curve = SECP256k1,
                     hash_function: HashFunctionType = sha3_256
                     ) -> KeyType:
    return key_cls.from_string(hex_bytes,
                               curve=curve,
                               hashfunc=hash_function)


hex_bytes_to_signing_key = partial(hex_bytes_to_key,
                                   SigningKey)
hex_bytes_to_verifying_key = partial(hex_bytes_to_key,
                                     VerifyingKey)


def load_signing_key(key_json: Dict[str, Union[str, int, dict]],
                     password: str) -> bytes:
    try:
        crypto = key_json['crypto']
    except KeyError:
        crypto = key_json['Crypto']

    aes_key, head_mac_bytes = derived_key_parts(password=password,
                                                kdf_params=crypto['kdfparams'])
    cipher_text = decode_hex(crypto['ciphertext'])
    validate_mac(cipher_text=cipher_text,
                 head_mac_bytes=head_mac_bytes,
                 predicted_mac=decode_hex(crypto['mac']))

    cipher_params = crypto['cipherparams']
    initial_value = initial_value_str_to_int(string=cipher_params['iv'])
    counter = Counter.new(nbits=128,
                          initial_value=initial_value,
                          allow_wraparound=True)
    return aes.decrypt(text=cipher_text,
                       key=aes_key,
                       counter=counter,
                       mode=AES.MODE_CTR)


def derived_key_parts(*,
                      password: str,
                      kdf_params: Dict[str, Union[int, str]]
                      ) -> Tuple[bytes, bytes]:
    derived_key = scrypt_hash(password,
                              salt=decode_hex(kdf_params['salt']),
                              buffer_length=kdf_params['dklen'],
                              p=kdf_params['p'],
                              r=kdf_params['r'],
                              n=kdf_params['n'])
    if len(derived_key) < 32:
        err_msg = 'Derived key must be at least 32 bytes long.'
        raise ValueError(err_msg)
    return derived_key[:16], derived_key[16:32]


def validate_mac(*,
                 cipher_text: bytes,
                 head_mac_bytes: bytes,
                 predicted_mac: bytes) -> None:
    mac_bytes = head_mac_bytes + cipher_text
    mac = keccak_256_hash(mac_bytes).digest()
    if mac != predicted_mac:
        err_msg = ('MAC mismatch. '
                   'Password seems to be incorrect.')
        raise ValueError(err_msg)


def initial_value_str_to_int(string: str) -> int:
    return big_endian_to_int(decode_hex(string))



import hmac
from hashlib import sha256
from typing import (Callable,
                    Tuple)

from ecdsa import SECP256k1

from .coding import (encode_number,
                     decode_number,
                     hex_string_to_int)
from .coordinates import affine
from .coordinates.utils import modular_multiplicative_inverse
from .hashes import keccak_256_hash
from .restrictions import (R_LENGTH,
                           S_LENGTH)
from .types import AffinePointType

secp256k_generator_point = SECP256k1.generator
secp256k_generator_point = (secp256k_generator_point.x(),
                            secp256k_generator_point.y())


def sign_message(message: str,
                 *,
                 signing_key_hex_string: str) -> str:
    message_hash = hash_message(message)
    v, r, s = signature_triplet(message_hash=message_hash,
                                signing_key_hex_string=signing_key_hex_string)
    return hex(r)[2:].zfill(R_LENGTH) + hex(s)[2:].zfill(S_LENGTH) + hex(v)[2:]


def add_prefix(message: str) -> str:
    # more info at
    # https://github.com/ethereum/go-ethereum/commit/b59c8399fbe42390a3d41e945d03b1f21c1a9b8d
    return ("\x19Ethereum Signed Message:\n"
            + str(len(message))
            + message)


def hash_message(message: str,
                 *,
                 message_modifier: Callable[[str], str] = add_prefix,
                 encoding: str = 'utf-8') -> str:
    modified_message = message_modifier(message)
    encoded_message = modified_message.encode(encoding)
    return keccak_256_hash(encoded_message).hexdigest()


def signature_triplet(
        *,
        message_hash: str,
        signing_key_hex_string: str,
        generator_point: AffinePointType = secp256k_generator_point,
        order: int = SECP256k1.order) -> Tuple[int, int, int]:
    message_hash_int = hex_string_to_int(message_hash)
    signing_key_int = hex_string_to_int(signing_key_hex_string)

    k = deterministic_generate_k(message_hash=message_hash,
                                 signing_key_hex_string=signing_key_hex_string)
    r, y = affine.multiply(point=generator_point,
                           multiplier=k)
    s = (modular_multiplicative_inverse(k, order)
         * (message_hash_int + r * signing_key_int) % order)
    v = 27 + ((y % 2) ^ (0 if s * 2 < order else 1))
    s = s if s * 2 < order else order - s
    return v, r, s


def deterministic_generate_k(*,
                             message_hash: str,
                             signing_key_hex_string: str) -> int:
    signing_key_int = hex_string_to_int(signing_key_hex_string)
    message_hash_int = hex_string_to_int(message_hash)
    signing_key_bytes = encode_number(signing_key_int,
                                      base=256,
                                      min_length=32)
    message_hash_bytes = encode_number(message_hash_int,
                                       base=256,
                                       min_length=32)

    v = b'\x01' * 32
    key = b'\x00' * 32
    key = hmac.new(key=key,
                   msg=v + b'\x00' + signing_key_bytes + message_hash_bytes,
                   digestmod=sha256).digest()
    v = hmac.new(key=key,
                 msg=v,
                 digestmod=sha256).digest()
    key = hmac.new(key=key,
                   msg=v + b'\x01' + signing_key_bytes + message_hash_bytes,
                   digestmod=sha256).digest()
    v = hmac.new(key=key,
                 msg=v,
                 digestmod=sha256).digest()
    return decode_number(hmac.new(key=key,
                                  msg=v,
                                  digestmod=sha256).digest(),
                         base=256)



import codecs
import string
from functools import partial
from typing import Union

characters_by_bases = {
    2: '01',
    8: string.octdigits,
    10: string.digits,
    16: string.digits + 'abcdef',
    32: string.ascii_lowercase + '234567',
    58: '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz',
    256: ''.join(map(chr, range(256)))
}


def encode_number(number: int,
                  *,
                  base: int,
                  min_length: int) -> bytes:
    characters = characters_by_base(base)
    result = bytes()
    while number > 0:
        character = characters[number % base]
        result = bytes([ord(character)]) + result
        number //= base

    pad_size = min_length - len(result)

    padding_element = b'\x00' if base == 256 else b'1' if base == 58 else b'0'
    if pad_size > 0:
        result = padding_element * pad_size + result

    return result


def decode_number(bytes_or_string: Union[str, bytes],
                  *,
                  base: int) -> int:
    if base == 16:
        bytes_or_string = bytes_or_string.lower()

    if base == 256 and isinstance(bytes_or_string, str):
        bytes_or_string = bytes(bytearray.fromhex(bytes_or_string))

    if base == 256:
        def position(byte: int) -> int:
            return byte
    else:
        characters = characters_by_base(base)
        characters_positions = {position: character
                                for (position,
                                     character) in enumerate(characters)}

        def position(byte_or_character: Union[int, str]) -> int:
            char = (byte_or_character
                    if isinstance(byte_or_character, str)
                    else chr(byte_or_character))
            return characters_positions[char]

    result = 0
    for byte_or_character in bytes_or_string:
        result *= base
        result += position(byte_or_character)
    return result


def characters_by_base(base: int) -> str:
    try:
        return characters_by_bases[base]
    except KeyError:
        raise ValueError(f'Invalid base: {base}.')


def hex_string_to_int(hex_string: str) -> int:
    return int(hex_string, 16)


def big_endian_to_int(big_endian: bytes) -> int:
    return int.from_bytes(big_endian.lstrip(bytes([0])),
                          byteorder='big')


decode_hex = partial(codecs.decode,
                     encoding='hex')



import pytest
from ethereum_helpers.messages import sign_message
from tests.strategies import messages_strategy
from tests.utils import example


@pytest.fixture(scope='function')
def message() -> str:
    return example(messages_strategy)


@pytest.fixture(scope='function')
def signature(message: str,
              signing_key_hex_string: str) -> str:
    return sign_message(message,
                        signing_key_hex_string=signing_key_hex_string)



from ecdsa import (SigningKey,
                   VerifyingKey)

from .hashes import keccak_256_hash


def signing_key_to_address(key: SigningKey) -> str:
    # based on
    # https://www.reddit.com/r/ethereum/comments/69qfkv/python_script_that_converts_ethereum_private_key/
    verifying_key = key.get_verifying_key()
    return verifying_key_to_address(verifying_key)


def verifying_key_to_address(key: VerifyingKey) -> str:
    key_string = key.to_string()
    key_hash = keccak_256_hash(key_string)
    key_hash_hex = key_hash.hexdigest()
    return key_hash_hex[24:]



from ethereum_helpers.messages import sign_message


def test_sign_message(hex_characters: str,
                      message: str,
                      signing_key_hex_string: str) -> None:
    signature = sign_message(message,
                             signing_key_hex_string=signing_key_hex_string)
    assert isinstance(signature, str)
    assert not signature.startswith('0x')
    assert all(character in hex_characters
               for character in signature)



from codecs import encode
from hashlib import sha256

import pytest
from ecdsa import (SigningKey,
                   VerifyingKey)
from ecdsa.curves import (SECP256k1,
                          Curve)
from ethereum_helpers.keys import HashFunctionType


@pytest.fixture(scope='function')
def curve() -> Curve:
    return SECP256k1


@pytest.fixture(scope='function')
def hash_function() -> HashFunctionType:
    return sha256


@pytest.fixture(scope='function')
def signing_key(curve: Curve,
                hash_function: HashFunctionType) -> SigningKey:
    return SigningKey.generate(curve=curve,
                               hashfunc=hash_function)


@pytest.fixture(scope='function')
def verifying_key(signing_key: SigningKey) -> VerifyingKey:
    return signing_key.get_verifying_key()


@pytest.fixture(scope='function')
def signing_key_hex_bytes(signing_key: SigningKey) -> bytes:
    return signing_key.to_string()


@pytest.fixture(scope='function')
def verifying_key_hex_bytes(verifying_key: VerifyingKey) -> bytes:
    return verifying_key.to_string()


@pytest.fixture(scope='function')
def signing_key_hex_string(signing_key_hex_bytes: bytes) -> bytes:
    return encode(signing_key_hex_bytes, 'hex').decode('ascii')


@pytest.fixture(scope='function')
def verifying_key_hex_string(verifying_key_hex_bytes: bytes) -> bytes:
    return encode(verifying_key_hex_bytes, 'hex').decode('ascii')
